
# grammar.py

from __future__ import print_function

from lexer import Lexer

# --------------------------------------------------------------------------

class Rule (object) :
   def __init__ (self) :
       self.name = ""
       self.expr = None
       self.rule_mode = ""
       self.rule_type = ""
       self.super_type = ""
       # self.nullable = False
       # self.first = [ ]
       # self.line = 1

class Expression (object) :
   def __init__ (self) :
       self.alternatives = [ ]
       # self.nullable = False
       # self.first = [ ]
       # self.line = 1

class Alternative (object) :
   def __init__ (self) :
       self.items = [ ]
       self.silent = False
       # self.nullable = False
       # self.first = [ ]
       # self.line = 1

class Ebnf (object) :
   def __init__ (self) :
       self.mark = ""
       self.expr = None
       self.impl = None
       # self.nullable = False
       # self.first = [ ]
       # self.line = 1

class Nonterminal (object) :
   def __init__ (self) :
       self.variable = ""
       self.rule = ""
       self.add = False
       # self.call
       # self.symbol = None
       # self.nullable = False
       # self.first = [ ]
       # self.line = 1

class Terminal (object) :
   def __init__ (self) :
       self.text = ""
       # self.symbol = None
       # self.nullable = False
       # self.first = [ ]
       # self.line = 1

# --------------------------------------------------------------------------

class Directive (object) :
   def __init__ (self) :
       self.name = ""

class Assign (Directive) :
   def __init__ (self) :
       self.variable = ""
       self.value = ""

class Style (Directive) :
   def __init__ (self) :
       self.name = ""

# --------------------------------------------------------------------------

class Options (object) :
   def __init__ (self) :
       self.store_name = ""
       self.store_value = ""
       self.add = False
       self.call = False
       self.need_nonterminal = False

# --------------------------------------------------------------------------

class Grammar (Lexer) :

   def __init__ (self) :
       super (Grammar, self).__init__ ()
       self.rules = [ ]
       #
       # self.keywords = { }
       # self.delimiters = { }
       # self.special_list = [ ]
       # self.special_dict = { }
       #
       # self.symbol_cnt = 0
       # self.symbols = [ ]
       # self.symbol_dict = { }
       #
       # self.rule_dict = { }
       # self.nullableChanged = True
       # self.firstChanged = True
       #
       # self.collections = { }

   def skipDirective (self) :
       while self.isSeparator ('<') :
          self.nextToken ()
          while not self.isEndOfSource () and not self.isSeparator ('>') :
             self.nextToken ()
          if self.isEndOfSource () :
             self.error ("Unfinished directive")
          self.nextToken () # skip '>'

   def readIdentifier (self, msg) :
       name = ""
       if not self.isIdentifier () :
          self.error (msg)
       else :
          name = self.tokenText
          self.nextToken ()
       return name

   def ruleDirective (self, rule) :
       if self.isSeparator ('<') :
          self.nextToken ()
          if self.isKeyword ("new") :
             self.nextToken ()
             rule.rule_mode = "new"
          elif self.isKeyword ("modify") :
             self.nextToken ()
             rule.rule_mode = "modify"
          elif self.isKeyword ("return") :
             self.nextToken ()
             rule.rule_mode = "return"
          elif self.isKeyword ("select") :
             self.nextToken ()
             rule.rule_mode = "select"
          elif self.isKeyword ("choose") :
             self.nextToken ()
             rule.rule_mode = "choose"
          elif self.isKeyword ("arit") :
             self.nextToken ()
             rule.rule_mode = "arit"

          rule.rule_type = self.readIdentifier ("Type name expected")

          if self.isSeparator (':') :
             self.nextToken ()
             rule.super_type = self.readIdentifier ("Super-type expected")

       self.checkSeparator ('>')

       if self.isSeparator ('<') :
          self.nextToken ()
          if self.isKeyword ("custom") :
             self.nextToken ()
             self.readIdentifier ("")
             while self.isSeparator (',') :
                self.nextToken ()
                self.readIdentifier ("")
          self.checkSeparator ('>')

   def itemDirective (self, alt, opt) :
       while self.isSeparator ('<') :
          self.nextToken ()

          if self.isKeyword ("store") :
             self.nextToken ()
             opt.store_name = self.readIdentifier ("Field identifier expected")
             self.checkSeparator (':')
             opt.store_type = self.readIdentifier ("Type identifier expected")
          elif self.isKeyword ("add") :
             self.nextToken ()
             opt.add = True
             opt.need_nonterminal = True
          elif self.isKeyword ("call") :
             self.nextToken ()
             opt.call = True
             opt.need_nonterminal = True
          elif self.isKeyword ("set") :
             self.nextToken ()
             variable = self.readIdentifier ("Variable identifier expected")
             self.checkSeparator ('=')
             value = self.readIdentifier ("Value expected")
             # add Assign to this alternative
             item = Assign ()
             item.variable = variable
             item.value = value
             alt.items.append (item)
          elif self.isKeyword ("silent") :
             self.nextToken ()
             # set silent field in this alternative
             alt.silent = True
          elif ( self.isKeyword ("indent") or
                 self.isKeyword ("unindent") or
                 self.isKeyword ("no_space") or
                 self.isKeyword ("new_line") or
                 self.isKeyword ("empty_line") ) :
             name = self.tokenText
             self.nextToken ()
             # add Style to this alternative
             item = Style ()
             item.name = name
             alt.items.append (item)
          else :
             self.error ("Unknown directive " + self.tokenText)

          self.checkSeparator ('>')

   def setPosition (self, item) :
       item.line = self.tokenLineNum

   # -- rules --

   def parseRules (self) :
       while not self.isEndOfSource () :
          self.parseRule ()

   def parseRule (self) :
       rule = Rule ()

       if self.token != self.identifier :
          self.error ("Rule idenitifier expected")
       rule.name = self.tokenText
       self.setPosition (rule)
       self.nextToken ()

       # self.skipDirective ()
       self.ruleDirective (rule)

       self.checkSeparator (':')

       if rule.rule_mode == "select" :
          rule.expr = self.selectExpression ()
       elif rule.rule_mode == "choose" :
          rule.expr = self.chooseExpression ()
       elif rule.rule_mode == "arit" :
          rule.expr = self.aritExpression ()
       else :
          rule.expr = self.parseExpression ()
       self.checkSeparator (';')

       self.rules.append (rule)

   # -- normal expression --

   def parseExpression (self) :
       expr = Expression ()
       self.setPosition (expr)
       expr.alternatives.append (self.parseAlternative ())
       while self.isSeparator ('|') :
          self.nextToken ()
          expr.alternatives.append (self.parseAlternative ())
       return expr

   def parseAlternative (self) :
       alt = Alternative ()
       self.setPosition (alt)
       # self.skipDirective ()
       opt = Options ()
       self.itemDirective (alt, opt)
       while not self.isSeparator ('|') and not self.isSeparator (')') and not self.isSeparator (';') :
          if self.token == self.identifier :
             item = self.parseNonterminal (opt)
             alt.items.append (item)
          elif self.token == self.character or self.token == self.string :
             if opt.need_nonterminal :
                self.error ("Nonterminal expected")
             item = self.parseTerminal ()
             alt.items.append (item)
          elif self.isSeparator ('(') :
             if opt.need_nonterminal :
                self.error ("Nonterminal expected")
             item = self.parseEbnf ()
             alt.items.append (item)
          else:
             self.error ("Unknown grammar item")
          # self.skipDirective ()
          opt = Options ()
          self.itemDirective (alt, opt)
       if opt.need_nonterminal :
          self.error ("Nonterminal expected")
       return alt

   def parseEbnf (self) :
       item = Ebnf ()
       self.setPosition (item)
       self.checkSeparator ('(')
       item.expr = self.parseExpression ()
       self.checkSeparator (')')

       if self.isSeparator ('?') :
          item.mark = '?'
          self.nextToken ()
       elif self.isSeparator ('+') or self.isSeparator ('@') :
          item.mark = '+'
          self.nextToken ()
       elif self.isSeparator ('*') :
          item.mark = '*'
          self.nextToken ()
       elif self.isSeparator ('=') :
          self.nextToken ();
          self.checkSeparator ('>')
          item.impl = self.parseAlternative ()

       return item

   def parseNonterminal (self, opt) :
       item = Nonterminal ()
       self.setPosition (item)
       item.add = opt.add
       item.call = opt.call

       if self.token != self.identifier :
          self.error ("identifier expected")
       name = self.tokenText
       self.nextToken ()

       if not self.isSeparator (':') :
          item.rule = name
       else :
          if item.add or item.call :
             self.error ("Variable identifier not allowed after <add> or <call> directive")
          self.nextToken () # skip ':'
          if not self.token == self.identifier :
             self.error ("rule identifier expected")
          item.variable = name
          item.rule = self.tokenText
          self.nextToken ()

       return item

   def parseTerminal (self) :
       item = Terminal ()
       self.setPosition (item)
       if self.token != self.character and self.token != self.string :
          self.error ("string expected")
       item.text = self.tokenText
       self.nextToken ()
       return item

   # -- select expression --

   " rule | rule | rule "

   def selectExpression (self) :
       expr = Expression ()
       self.setPosition (expr)
       expr.alternatives.append (self.selectAlternative ())
       while self.isSeparator ('|') :
          self.nextToken ()
          expr.alternatives.append (self.selectAlternative ())
       return expr

   def selectAlternative (self) :
       alt = Alternative ()
       self.setPosition (alt)
       item = self.selectNonterminal ()
       alt.items.append (item)
       return alt

   def selectNonterminal (self) :
       item = Nonterminal ()
       self.setPosition (item)
       if self.token != self.identifier :
          self.error ("identifier expected")
       item.rule = self.tokenText
       self.nextToken ()
       return item

   # -- choose expression --

   " rule ( rule | rule )* "

   def chooseExpression (self) :
       expr = Expression ()
       self.setPosition (expr)
       expr.alternatives.append (self.chooseAlternative ())
       return expr

   def chooseAlternative (self) :
       alt = Alternative ()
       self.setPosition (alt)

       item = self.chooseNonterminal ()
       alt.items.append (item)

       item = self.chooseEbnf ()
       alt.items.append (item)
       return alt

   def chooseNonterminal (self) :
       item = Nonterminal ()
       self.setPosition (item)
       if self.token != self.identifier :
          self.error ("identifier expected")
       item.rule = self.tokenText
       self.nextToken ()
       return item

   def chooseEbnf (self) :
       item = Ebnf ()
       self.setPosition (item)
       self.checkSeparator ('(')
       item.expr = self.chooseExpression2 ()
       self.checkSeparator (')')
       if self.isSeparator ('*') :
          item.mark = '*'
          self.nextToken ()
       return item

   def chooseExpression2 (self) :
       expr = Expression ()
       self.setPosition (expr)
       expr.alternatives.append (self.chooseAlternative2 ())
       while self.isSeparator ('|') :
          self.nextToken ()
          expr.alternatives.append (self.chooseAlternative2 ())
       return expr

   def chooseAlternative2 (self) :
       alt = Alternative ()
       self.setPosition (alt)
       item = self.chooseNonterminal ()
       alt.items.append (item)
       return alt

   # -- arit expression --

   " left:term  ( ( '+' <set fce=AddExp> | '-' <set fce=SubExp> ) right:simple_expr )? "

   def aritExpression (self) :
       expr = Expression ()
       self.setPosition (expr)
       expr.alternatives.append (self.aritAlternative ())
       return expr

   def aritAlternative (self) :
       alt = Alternative ()
       self.setPosition (alt)
       alt.items.append (self.aritNonterminal ()) # left:term
       alt.items.append (self.aritEbnf ())
       return alt

   def aritNonterminal (self) :
       item = Nonterminal ()
       self.setPosition (item)
       item.name = self.readIdentifier ("variable identifier expected")
       self.checkSeparator (':')
       item.rule = self.readIdentifier ("rule identifier expected")
       return item

   def aritEbnf (self) :
       item = Ebnf ()
       self.setPosition (item)
       self.checkSeparator ('(')
       item.expr = self.aritExpression2 ()
       self.checkSeparator (')')
       if self.isSeparator ('?') :
          item.mark = '?'
          self.nextToken ()
       elif self.isSeparator ('*') :
          item.mark = '*'
          self.nextToken ()
       return item

   " ( '+' <set fce=AddExp> | '-' <set fce=SubExp> ) right:simple_expr "

   def aritExpression2 (self) :
       expr = Expression ()
       self.setPosition (expr)
       expr.alternatives.append (self.aritAlternative2 ())
       return expr

   def aritAlternative2 (self) :
       alt = Alternative ()
       self.setPosition (alt)
       alt.items.append (self.aritEbnf2 ()) # ( '+' <set fce=AddExp> | '-' <set fce=SubExp> )
       alt.items.append (self.aritNonterminal ()) # right:simple_expr
       return alt

   def aritEbnf2 (self) :
       item = Ebnf ()
       self.setPosition (item)
       self.checkSeparator ('(')
       item.expr = self.aritExpression3 ()
       self.checkSeparator (')')
       return item

   " '+' <set fce=AddExp> | '-' <set fce=SubExp> "

   def aritExpression3 (self) :
       expr = Expression ()
       self.setPosition (expr)
       expr.alternatives.append (self.aritAlternative3 ())
       while self.isSeparator ('|') :
          self.nextToken ()
          expr.alternatives.append (self.aritAlternative3 ())
       return expr

   def aritAlternative3 (self) :
       alt = Alternative ()
       self.setPosition (alt)
       alt.items.append (self.parseTerminal ()) # '-'
       alt.items.append (self.aritAssign ()) #  <set fce=SubExp>
       return alt

   def aritAssign (self) :
       self.checkSeparator ('<')
       self.checkKeyword ("set")

       item = Assign ()
       item.variable = self.readIdentifier ("Variable identifier expected")
       self.checkSeparator ('=')
       item.value = self.readIdentifier ("Value expected")

       self.checkSeparator ('>')
       return item

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    grammar = Grammar ()
    grammar.openFile ("pas.g")
    grammar.parseRules ()
    for rule in grammar.rules :
       print (rule.name)

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
