
# lexer.py

from __future__ import print_function
import os

# --------------------------------------------------------------------------

fileNames = [ ]

class Separators (object) :
   def __init__ (self) :
       self.value = -1 # -1 ... substring is not a valid symbol
       self.items = { } # another symbols with same prefix

class Define (object) :
   def __init__ (self) :
       self.name = ""
       self.params = None # None ... without paramaters, [ ] ... with empty ( )
       self.code = ""

class CondSection (object) :
   def __init__ (self) :
      self.orig_use_input = False # outer use_input
      self.already_used = False # already one sub-section used for token generation
      self.allow_else = True # False => "else" already found

class LexerException (Exception) :
   pass

# --------------------------------------------------------------------------

class LexerInput (object) :
   def __init__ (self) :
       self.source = ""

       self.byteOfs = 0 # byte offset (from 0)
       self.fileInx = 0 # file index
       self.lineNum = 1 # line number (from 1)
       self.colNum  = 1 # column number (from 1)

       # self.lastChar = ' '

       self.isInclude = False
       self.isMacro = False

       self.macros = { }
       self.above = None

   # -----------------------------------------------------------------------

   def readChar (self) :
       if self.byteOfs < self.sourceLen :
          self.ch = self.source [self.byteOfs]
          self.byteOfs = self.byteOfs + 1
          if self.ch == '\n' :
             self.lineNum = self.lineNum + 1
             self.colNum = 0
          elif self.ch != '\r' :
             self.colNum = self.colNum + 1
       else :
          self.ch = '\0'

   # -----------------------------------------------------------------------

   def setFileName (self, name) :
       global fileNames
       if name not in fileNames :
          fileNames.append (name)
       self.fileInx = fileNames.index (name) + 1

   def getFileName (self) :
       global fileNames
       if self.fileInx == 0 :
          return ""
       else :
          return fileNames [self.fileInx - 1]

   # -----------------------------------------------------------------------

   def openFile (self, fileName) :
       fileName = os.path.abspath (fileName)
       self.setFileName (fileName)
       f = open (fileName, "r")
       self.source = f.read ()
       self.sourceLen = len (self.source)

   def openString (self, sourceText) :
       self.source = str (sourceText) # optionaly conversion from QString
       self.sourceLen = len (self.source)

   def getSource (self) :
       return self.source # !?

# --------------------------------------------------------------------------

class Lexer (object) :
   eos = 0
   identifier = 1
   number = 2
   real = 3
   character = 4
   string = 5
   separator = 6
   end_of_line = 7

   def __init__ (self) :
       self.pascal = False
       self.lexer_keywords = None
       self.lexer_separators = None

       self.at_keywords = False
       self.tolerate_missing_include = False
       self.follow = None

       self.reset ()

   def reset (self) :

       self.charByteOfs = 0 # character (in self.ch)
       self.charFileInx = 0
       self.charLineNum = 1
       self.charColNum = 1

       self.tokenByteOfs = 0 # token (in self.token)
       self.tokenFileInx = 0
       self.tokenLineNum = 1
       self.tokenColNum = 1

       self.ch = '\0'
       self.token = self.eos
       self.tokenText = ""

       self.input = None
       self.input_stack = [ ]

       self.inside_directive = False
       self.initPreproc ()

   # -----------------------------------------------------------------------

   def openFile (self, fileName) :
       # self.reset ()
       self.input = LexerInput ()
       self.input.openFile (fileName)
       if self.follow : self.follow.start (self)
       self.nextChar ()
       self.nextToken ()

   def openString (self, sourceText) :
       # self.reset ()
       self.input = LexerInput ()
       self.input.openString (sourceText)
       if self.follow : self.follow.start (self)
       self.nextChar ()
       self.nextToken ()

   # -----------------------------------------------------------------------

   def indexToFileName (self, inx) :
       global fileNames
       if inx == 0 :
          return ""
       else :
          return fileNames [inx - 1]

   def fileNameToIndex (self, name) :
       global fileNames
       if not name in fileNames :
          return ""
       else :
          return fileNames.index (name) + 1

   def getFileName (self) :
       return self.input.getFileName () # !?

   def getSource (self) :
       return self.input.source # !?

   # -----------------------------------------------------------------------

   def warning (self, text) :
       print (self.input.getFileName () + ":" + str (self.charLineNum) + ":" + str (self.charColNum) + ": warnig: " + text)

   def error (self, text) :
       raise LexerException (self.input.getFileName () + ":" + str (self.charLineNum) + ":" + str (self.charColNum) + ": error: " + text)

   # -----------------------------------------------------------------------

   def nextChar (self) :
       self.charByteOfs = self.input.byteOfs
       self.charFileInx = self.input.fileInx
       self.charLineNum = self.input.lineNum
       self.charColNum  = self.input.colNum

       self.input.readChar ()
       self.ch = self.input.ch

       if self.follow: self.follow.oneCharacter ()

   def repeatChar (self) :
       self.charByteOfs = self.input.byteOfs
       self.charFileInx = self.input.fileInx
       self.charLineNum = self.input.lineNum
       self.charColNum  = self.input.colNum

       self.ch = self.input.ch

   # -----------------------------------------------------------------------

   def mark (self) :
       return (self.tokenByteOfs, self.tokenFileInx, self.tokenLineNum, self.tokenColNum)

   def rewind (self, params) :
       self.byteOfs = params [0]
       self.fileInx = params [1] # !?
       self.lineNum = params [2]
       self.colNum = params [3]
       self.nextChar ()
       self.nextToken ()

   # -----------------------------------------------------------------------

   def isLetter (self, c) :
       return c >= 'A' and c <= 'Z' or c >= 'a' and c <= 'z' or c == '_'

   def isDigit (self, c) :
       return c >= '0' and c <= '9'

   def isOctDigit (self, c) :
       return c >= '0' and c <= '7'

   def isHexDigit (self, c) :
       return c >= 'A' and c <= 'F' or c >= 'a' and c <= 'f' or c >= '0' and c <= '9'

   def isLetterOrDigit (self, c) :
       return self.isLetter (c) or self.isDigit (c)

   # -----------------------------------------------------------------------

   def comment_eol (self, delta = 2) :
       if self.follow : self.follow.startComment (2)

       while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' :
          self.nextChar ()

       if self.follow: self.follow.comment ()

   def comment (self, mark1) :
       if self.follow : self.follow.startComment (1)

       while self.ch != '\0' and self.ch != mark1 :
          if trace:
             self.tokenText = self.tokenText + self.ch
          self.nextChar ()
       if self.ch == mark1 :
          self.nextChar () # skip mark1
       else :
          self.error ("Unterminated comment")

       if self.follow: self.follow.comment ()

   def comment2 (self, mark1, mark2) :
       if self.follow : self.follow.startComment (2)

       prev = ' '
       while self.ch != '\0' and not (prev == mark1 and self.ch == mark2) :
          prev = self.ch
          self.nextChar ()
       if self.ch == mark2 :
          self.nextChar () # skip mark2
       else :
          self.error ("Unterminated comment")

       if self.follow : self.follow.comment ()

   def checkDigit (self) :
       if not self.isDigit (self.ch) :
          self.error ("Digit expected")

   def digits (self) :
       while self.isDigit (self.ch) :
          self.tokenText = self.tokenText + self.ch
          self.nextChar ()

   def stringChar (self) :
       if self.ch != '\\' :
          c = self.ch
          self.nextChar ()
          return c
       else :
          self.nextChar () # skip backslash

          if self.ch >= '0' and self.ch <= '7' :
             cnt = 1
             c = 0
             while self.ch >= '0' and self.ch <= '7' and cnt <= 3 :
                c = 8 * c  + ord (self.ch) - ord ('0')
                cnt = cnt + 1
                self.nextChar ()
             return chr (c)

          elif self.ch == 'x' or self.ch <= 'X' :
             self.nextChar ()
             c = 0
             while self.ch >= '0' and self.ch <= '9' or self.ch >= 'A' and self.ch <= 'Z' or self.ch >= 'a' and self.ch <= 'z':
                if self.ch >= '0' and self.ch <= '9' :
                   n = ord (self.ch) - ord ('0')
                elif self.ch >= 'A' and self.ch <= 'Z' :
                   n = ord (self.ch) - ord ('A') + 10
                else :
                   n = ord (self.ch) - ord ('a') + 10
                c = 16 * c  + n
                self.nextChar ()
             return chr (c % 255)

          elif self.ch == 'a' :
             c = '\a'
          elif self.ch == 'b' :
             c =  '\b'
          elif self.ch == 'f' :
             c =  '\f'
          elif self.ch == 'n' :
             c =  '\n'
          elif self.ch == 'r' :
             c =  '\r'
          elif self.ch == 't' :
             c =  '\t'
          elif self.ch == 'v' :
             c =  '\v'
          elif self.ch == '\'' or self.ch == '\"'  or ch == '?' :
             c = self.ch
          else :
             c = self.ch # same

          self.nextChar ()
          return c

   def lookupKeyword (self) :
       if self.lexer_keywords != None :
          if self.tokenText in self.lexer_keywords :
             self.token = self.lexer_keywords [self.tokenText]

   def processSeparator (self) :
       # print ("separator start", self.tokenText)

       # self.tokenText ... first character
       if self.lexer_separators != None :
          cont = None
          if self.tokenText in self.lexer_separators :
             cont = self.lexer_separators [self.tokenText]

          stop = False
          while not stop :
             if cont != None and self.ch in cont.items :
                cont = cont.items [self.ch] # following map of separators
                self.tokenText = self.tokenText + self.ch # add one character
                # print ("separator character", self.ch)
                self.nextChar ()
             elif cont != None and cont.value >= 0 :
                # print ("separator", self.tokenText, "separator value ", cont.value)
                self.token = cont.value # new separator
                stop = True
             else :
                self.error ("Unknown separator")

   def backStep (self) :
       self.tokenByteOfs = self.tokenByteOfs - 1
       self.tokenColNum = self.tokenColNum - 1

   # -----------------------------------------------------------------------

   def normalToken (self) :
       self.token = self.eos
       self.tokenText = ""

       slash = False
       stop = False
       whiteSpace = True

       while whiteSpace and self.ch != '\0' and not stop:
          if self.inside_directive :
             while (self.ch != '\0' and self.ch <= ' ') or self.ch == '\\' :
                if self.ch == '\\' :
                   if self.follow: self.follow.whiteSpace ()
                   self.nextChar ()
                   if self.ch == '\r' :
                      if self.follow: self.follow.whiteSpace ()
                      self.nextChar ()
                   if self.ch == '\n' :
                      if self.follow: self.follow.whiteSpace ()
                      self.nextChar ()
                else :
                   if self.ch == '\r' or self.ch == '\n' :
                      stop = True
                   if self.follow: self.follow.whiteSpace ()
                   self.nextChar ()
          else :
             while self.ch != '\0' and self.ch <= ' ' :
                if self.follow: self.follow.whiteSpace ()
                self.nextChar ()

          whiteSpace = False
          if self.ch == '/' :
             self.nextChar () # skip '/'
             if self.ch == '/' :
                self.nextChar () # skip '/'
                self.comment_eol ()
                whiteSpace = True # check again for white space
             elif self.ch == '*' :
                self.nextChar () # skip '*'
                self.comment2 ('*', '/')
                whiteSpace = True # check again for white space
             else:
                slash = True # produce '/' token

       self.tokenByteOfs = self.charByteOfs
       self.tokenFileInx = self.charFileInx
       self.tokenLineNum = self.charLineNum
       self.tokenColNum = self.charColNum

       if slash :
          self.token = self.separator
          self.tokenText = '/'
          self.backStep ()
          self.processSeparator ()

       elif stop :
          self.token = self.end_of_line

       elif self.ch == '\0' :
          self.token = self.eos

       elif self.isLetter (self.ch) :
          self.token = self.identifier
          while self.isLetterOrDigit (self.ch) :
             self.tokenText = self.tokenText + self.ch
             self.nextChar ()
          self.lookupKeyword ()

       elif self.ch == '@' and self.at_keywords :
          self.token = self.identifier
          while self.isLetterOrDigit (self.ch) or self.ch == '@':
             self.tokenText = self.tokenText + self.ch
             self.nextChar ()
          self.lookupKeyword ()

       elif self.isDigit (self.ch) :
          self.token = self.number
          cont = True
          if self.ch == "0" :
             self.tokenText = self.tokenText + self.ch
             if self.ch == "x" :
                while self.isHexDigit (self.ch) :
                   self.tokenText = self.tokenText + self.ch
                   self.nextChar ()
                cont = False
             else :
                while self.isOctDigit (self.ch) :
                   self.tokenText = self.tokenText + self.ch
                   self.nextChar ()
                cont = False
          if cont :
             self.digits ()
             if self.ch == '.' :
                self.token = self.real
                self.tokenText = self.tokenText + self.ch # store '.'
                self.nextChar () # skip '.'
                self.checkDigit ()
                self.digits ()
             if self.ch == 'e' or self.ch == 'E':
                self.token = self.real
                self.tokenText = self.tokenText + self.ch # store 'e'
                self.nextChar () # skip 'e'
                if self.ch == '+' or self.ch == '-':
                   self.tokenText = self.tokenText + self.ch # store '+' or '-'
                   self.nextChar () # skip '+' or '-'
                self.checkDigit ()
                self.digits ()
             if self.ch == 'l' or self.ch == 'L':
                self.nextChar ()

       elif self.ch == '\'' :
          self.token = self.character
          self.nextChar ()
          while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch != '\'' :
             self.tokenText = self.tokenText + self.stringChar ()
          if self.ch != '\'' :
             self.error ("Unterminated string")
          self.nextChar ()

       elif self.ch == '\"' :
          self.token = self.string
          self.nextChar ()
          while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch != '\"' :
             self.tokenText = self.tokenText + self.stringChar ()
          if self.ch != '\"' :
             self.error ("Unterminated string")
          self.nextChar ()

       else :
          self.token = self.separator
          self.tokenText = self.ch
          self.nextChar ()
          self.processSeparator ()

   # -----------------------------------------------------------------------

   def pascalToken (self) :
       self.token = self.eos
       self.tokenText = ""

       whiteSpace = True
       slash = False
       parenthesis = False

       while whiteSpace and self.ch != '\0':
          whiteSpace = False
          while self.ch != '\0' and self.ch <= ' ' :
             self.nextChar ()
          if self.ch == '/' :
             self.nextChar () # skip '/'
             if self.ch == '/' :
                self.nextChar () # skip '/'
                self.comment_eol ()
                whiteSpace = True # check again for white space
             else:
                slash = True # produce '/' token
          if self.ch == '{' :
             self.nextChar () # skip '{'
             self.comment ('}')
             whiteSpace = True # check again for white space
          if self.ch == '(' :
             self.nextChar () # skip '('
             if self.ch == '*' :
                self.nextChar () # skip '*'
                self.comment2 ('*', ')')
                whiteSpace = True # check again for white space
             else :
                parenthesis = True # produce '(' token

       self.tokenByteOfs = self.charByteOfs
       self.tokenFileInx = self.charFileInx
       self.tokenLineNum = self.charLineNum
       self.tokenColNum = self.charColNum

       if slash :
          self.token = self.separator
          self.tokenText = '/'
          self.backStep ()
          self.processSeparator ()

       elif parenthesis :
          self.token = self.separator
          self.tokenText = '('
          self.backStep ()
          self.processSeparator ()

       elif self.ch == '\0' :
          self.token = self.eos

       elif self.isLetter (self.ch) :
          self.token = self.identifier
          while self.isLetterOrDigit (self.ch) :
             self.tokenText = self.tokenText + self.ch
             self.nextChar ()
          self.tokenText = self.tokenText.lower ()
          self.lookupKeyword ()

       elif self.isDigit (self.ch) :
          self.token = self.number
          self.digits ()
          if self.ch == '.' :
             self.token = self.real
             self.tokenText = self.tokenText + self.ch # store '.'
             self.nextChar () # skip '.'
             self.checkDigit ()
             self.digits ()
          if self.ch == 'e' or self.ch == 'E':
             self.token = self.real
             self.tokenText = self.tokenText + self.ch # store 'e'
             self.nextChar () # skip 'e'
             if self.ch == '+' or self.ch == '-':
                self.tokenText = self.tokenText + self.ch # store '+' or '-'
                self.nextChar () # skip '+' or '-'
             self.checkDigit ()
             self.digits ()

       elif self.ch == '\'' :
          self.token = self.string
          self.nextChar ()
          again = True
          while again :
             again = False
             while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch != '\'' :
                self.tokenText = self.tokenText + self.ch
                self.nextChar ()
             if self.ch == '\'' :
                self.nextChar () # skip quote
                if self.ch == '\'' : # two quotes
                   self.tokenText = self.tokenText + self.ch # store one quote
                   self.nextChar ()
                   again = True
             else :
                self.error ("Unterminated string")

       else :
          self.token = self.separator
          self.tokenText = self.ch
          self.nextChar ()
          self.processSeparator ()

   # -----------------------------------------------------------------------

   def initPreproc (self) :
       self.incl_path = [ "." ]
       self.macros = { }

       self.use_input = True
       self.cond_stack = [ ]

   def addInclDir (self, path) :
       self.incl_path.append (path)

   def addMacro (self, name, value) :
       defn = Define ()
       defn.name = name
       defn.code = value
       self.macros [defn.name] = defn

   def searchFile (self, filename) :
       result = None
       for path in self.incl_path :
          test = os.path.join (path, filename)
          if os.path.exists (test):
             result = test
             # result = os.path.abspath (result)
             break
       return result

   def simpleToken (self) :
       if self.pascal :
          self.pascalToken ()
       else :
          self.normalToken ()

   def verifySeparator (self, value) :
       if not self.isSeparator (value) :
          self.error (value + " expected")

   def expandToken (self) :
       again = True
       while again :
          "read simple token"
          self.simpleToken ()
          # print ("simple token", self.tokenText)
          again = False
          if self.use_input and self.token == self.identifier :
             "look for macro definition"
             name = self.tokenText
             defn = None
             inp = self.input
             while defn == None and inp != None :
                if name in inp.macros :
                   defn = inp.macros [name]
                inp = inp.above
             if defn == None and name in self.macros :
                defn = self.macros [name]
             # print ("EXPAND", name, defn != None)

             if defn != None :
                "expand macro"
                if self.follow : self.follow.expandMacro (defn)
                # NO if self.follow : self.follow.paramToken ()
                next_input = LexerInput ()
                missing_parameters = False

                if defn.params != None :
                   self.simpleToken () # skip macro name
                   if not self.isSeparator ('(') :
                      missing_parameters = True
                   else :
                      if self.follow: self.follow.openParameters ()
                      if self.follow : self.follow.paramToken ()
                      self.simpleToken () # skip '('
                      cnt = len (defn.params)
                      for inx in range (cnt) :
                          param_name = defn.params [inx]
                          if self.follow: self.follow.openParam (param_name)
                          value = ""
                          while not self.isSeparator (',') and not self.isSeparator (')') :
                             if value != "" :
                                value = value + " "
                             value = value + self.tokenText
                             if self.follow : self.follow.paramToken ()
                             self.simpleToken ()
                          if self.follow: self.follow.closeParam ()
                          # print ("PARAM", param_name, value)
                          param = Define ()
                          param.name = param_name
                          param.code = value
                          next_input.macros [param.name] = param
                          if inx+1 < cnt :
                             self.verifySeparator (',')
                             if self.follow : self.follow.paramToken ()
                             self.simpleToken ()
                      self.verifySeparator (')')
                      if self.follow : self.follow.paramToken ()
                      # NO self.simpleToken ()
                      if self.follow: self.follow.closeParameters ()

                if not missing_parameters :
                   next_input.isMacro = True
                   next_input.above = self.input
                   # next_input.lastChar = self.ch
                   self.input_stack.append (self.input)
                   self.input = next_input
                   self.input.openString (defn.code)
                   self.nextChar ()
                   again = True

   def defineDirective (self) :
       defn = Define ()
       self.simpleToken () # skip define

       if not self.isIdentifier () :
          self.error ("Macro name expected")
       defn.name = self.tokenText
       with_params = (self.ch == '(') # no space before '(' allowed
       self.simpleToken () # skip name

       if with_params :
          defn.params = [ ]
          self.simpleToken () # skip '('
          while not self.isSeparator (')') :
             if not self.isIdentifier () :
                self.error ("Parameter name expected")
             defn.params.append (self.tokenText)
             self.simpleToken ()
             if not self.isSeparator (')') :
                self.verifySeparator (',')
                self.simpleToken ()
          self.simpleToken () # skip ')'

       while self.token != self.eos and self.token != self.end_of_line :
           defn.code = defn.code + self.tokenText + " "
           self.simpleToken ()

       # defn.code = self.tokenText
       # while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' :
       #    defn.code = defn.code + self.ch
       #    self.nextChar ()

       self.macros [defn.name] = defn
       if self.follow: self.follow.defineDirective (defn)

   def undefDirective (self) :
       self.simpleToken () # skip undef
       if not self.isIdentifier () :
          self.error ("Macro name expected")

       name = self.tokenText
       if name in self.macros :
          del self.macros [name]
       if self.follow: self.follow.undefDirective (name)

       self.simpleToken () # skip name

   def includeDirective (self) :
       self.simpleToken () # skip include
       if self.token == self.string :
          result = self.tokenText
       elif self.isSeparator ('<') :
          result = ""
          while self.ch != '\0' and self.ch != '\r' and self.ch != '\n' and self.ch != '>' :
             result = result + self.ch
             self.nextChar ()
          if self.ch != '>' :
             self.error ("Missing >")
          self.nextChar ()
       else :
          self.error ("Include file name expected")
       # self.simpleToken ()
       if not self.use_input :
          result = None
       return result

   def readMessage (self) :
       text = self.ch
       while self.ch != '\0' and self.ch != "\r" and self.ch != "\n" : # !? characters, not tokens
          if self.ch == '\\' :
             self.nextChar ()
             ok = False
             if self.ch == '\r' :
                self.nextChar ()
                ok = True
             if self.ch == '\n' :
                self.nextChar ()
                ok = True
             if not ok :
                text = text + '\\'
          else :
             text = text + self.ch
             self.nextChar ()
       self.token = self.end_of_line # !?
       return text

   def isDefined (self) :
       if not self.isIdentifier () :
          self.error ("Identifier expected")
       name = self.tokenText
       self.simpleToken ()
       return name in self.macros

   def readValue (self) :
       if self.isIdentifier () :
          if self.tokenText == "defined" :
             self.simpleToken ()
             if self.isSeparator ('(') :
                self.simpleToken ()
                value = self.isDefined ()
                self.verifySeparator (')')
                self.simpleToken ()
             else :
                value = self.isDefined ()
          else :
             name = self.tokenText
             if name in self.macros :
                text = self.macros [name].code
                text = text.strip ()
                if text.endswith ('l') or text.endswith ('L') :
                   text = text [:-1]
                value = int (text, 0) # also 0x
             else :
                # self.error ("Unknown identifier " + name)
                value = 0 # !?
             self.expandToken ()
       elif self.isNumber () :
          value = int (self.tokenText)
          self.expandToken ()
       elif self.isSeparator ('-') :
          self.expandToken ()
          value = - self.readValue ()
       elif self.isSeparator ('!') :
          self.expandToken ()
          value = not self.readValue ()
       elif self.isSeparator ('(') :
          self.expandToken ()
          value = self.expression ()
          self.verifySeparator (')')
          self.expandToken ()
       else :
          self.error ("Value expected " + self.tokenText)
       return value

   def readArit (self) :
       func = ""
       prio = 0
       if self.isSeparator ('+') or self.isSeparator ('-') :
          func = self.tokenText
          prio = 11
       elif self.isSeparator ('*') or self.isSeparator ('/') or self.isSeparator ('%') :
          func = self.tokenText
          prio = 12
       elif self.isSeparator ('&') :
          if self.ch == '&' :
             self.nextChar ()
             func = "&&"
             prio = 4
          else :
             func = self.tokenText
             prio = 7
       elif self.isSeparator ('|') :
          if self.ch == '|' :
             self.nextChar ()
             func = "||"
             prio = 3
          else :
             func = self.tokenText
             prio = 5
       elif self.isSeparator ('<') :
          if self.ch == '<' :
             self.nextChar ()
             func = "<<"
             prio = 10
          elif self.ch == '=' :
             self.nextChar ()
             func = "<="
             prio = 9
          else :
             func = self.tokenText
             prio = 9
       elif self.isSeparator ('>') :
          if self.ch == '>' :
             self.nextChar ()
             func = ">>"
             prio = 10
          elif self.ch == '=' :
             self.nextChar ()
             func = ">="
             prio = 9
          else :
             func = self.tokenText
             prio = 9
       elif self.isSeparator ('=') :
          if self.ch == '=' :
             self.nextChar ()
             func = "=="
             prio = 8
          else :
             self.error ("Unknown operator")
       elif self.isSeparator ('!') :
          if self.ch == '=' :
             self.nextChar ()
             func = "!="
             prio = 8
          else :
             self.error ("Unknown operator")
       else :
          self.error ("Unknown operator " + self.tokenText)
       self.expandToken ()
       return (func, prio)

   def calculate (self, a, func, b) :
       if func == "*" : return a * b
       elif func == "/" : return a / b
       elif func == "%" : return a % b
       elif func == "+" : return a + b
       elif func == "-" : return a - b
       elif func == "<<" : return a << b
       elif func == ">>" : return a >> b
       elif func == "<" : return a < b
       elif func == "<=" : return a <= b
       elif func == ">" : return a > b
       elif func == ">=" : return a >= b
       elif func == "==" : return a == b
       elif func == "!=" : return a != b
       elif func == "^" : return a ^ b
       elif func == "&" : return a & b
       elif func == "|" : return a | b
       elif func == "&&" : return a and b
       elif func == "||" : return a or b
       elif func == "," : return b
       else: self.error ("unknown operator " + func)

   def expression (self) :
       values = [ ]
       functions = [ ]
       priorities = [ ]

       value = self.readValue ()
       values.append (value)

       while self.token != self.eos and self.token != self.end_of_line and not self.isSeparator (')') :

          next_func, next_prio = self.readArit ()

          while len (priorities) > 0 and priorities [-1] >= next_prio :
             right = values.pop ()
             left = values.pop ()
             func = functions.pop ()
             priorities.pop ()
             value = self.calculate (left, func, right)
             values.append (value)

          functions.append (next_func)
          priorities.append (next_prio)

          value = self.readValue ()
          values.append (value)

       value = values.pop ()
       while len (priorities) > 0:
          left = values.pop ()
          func = functions.pop ()
          priorities.pop ()
          value = self.calculate (left, func, value)

       return value

   def ifDirective (self, symbol = False, negation = False) :
       if symbol :
          self.simpleToken ()
       else :
          self.expandToken ()
       if self.use_input :
          if symbol :
             cond = self.isDefined ()
          else :
             cond = bool (self.expression ())
          if negation :
             cond = not cond
       else :
          cond = False
       section = CondSection ()
       section.orig_use_input = self.use_input
       section.already_used = cond
       self.cond_stack.append (section)
       self.use_input = cond

   def elifDirective (self) :
       if len (self.cond_stack) == 0 :
          self.error ("Unexpected elif")
       section = self.cond_stack [-1]
       if not section.allow_else :
          self.error ("Unexpected elif after else")
       self.expandToken ()
       cond = bool (self.expression ())
       if not section.orig_use_input :
          self.use_input = False
       elif section.already_used :
          self.use_input = False
       else :
          self.use_input = cond
          if cond :
             section.already_used = True

   def elseDirective (self) :
       self.expandToken ()
       if len (self.cond_stack) == 0 :
          self.error ("Unexpected else")
       section = self.cond_stack [-1]
       if not section.allow_else :
          self.error ("Duplicated else")
       if not section.orig_use_input :
          self.use_input = False
       elif section.already_used :
          self.use_input = False
       else :
          self.use_input = True
          section.already_used = True
       section.allow_else = False

   def endifDirective (self) :
       self.simpleToken ()
       if len (self.cond_stack) == 0 :
          self.error ("Unexpected endif")
       section = self.cond_stack.pop ()
       self.use_input = section.orig_use_input

   def directive (self) :
       next_name = None
       self.inside_directive = True
       if self.follow: self.follow.startDirective ()
       self.simpleToken () # skip hash mark
       entry_use_input = self.use_input
       cond_directive = False

       if self.isKeyword ("define") :
          if self.use_input :
             self.defineDirective ()

       elif self.isKeyword ("undef") :
          if self.use_input :
             self.undefDirective ()

       elif self.isKeyword ("include") :
          if self.use_input :
             next_name = self.includeDirective ()

       elif self.isKeyword ("eval") :
          if self.use_input :
             self.expandToken ()
             value = self.expression ()
             if self.follow : self.follow.evalDirective (str (value))

       elif self.isKeyword ("if") :
          self.ifDirective ()
          cond_directive = True
          if self.follow : self.follow.ifDirective ()

       elif self.isKeyword ("ifdef") :
          self.ifDirective (symbol = True)
          cond_directive = True
          if self.follow : self.follow.ifdefDirective ()

       elif self.isKeyword ("ifndef") :
          self.ifDirective (symbol = True, negation = True)
          cond_directive = True
          if self.follow : self.follow.ifndefDirective ()

       elif self.isKeyword ("elif") :
          self.elifDirective ()
          cond_directive = True
          if self.follow : self.follow.elifDirective ()

       elif self.isKeyword ("else") :
          self.elseDirective ()
          cond_directive = True
          if self.follow : self.follow.elseDirective ()

       elif self.isKeyword ("endif") :
          self.endifDirective ()
          cond_directive = True
          if self.follow : self.follow.endifDirective ()

       elif self.isKeyword ("pragma") :
          if self.use_input :
             text = self.readMessage ()

       elif self.isKeyword ("warning") :
          if self.use_input :
             text = self.readMessage ()
             if self.follow : self.follow.warningDirective (text)
             self.warning (text)

       elif self.isKeyword ("error") :
          if self.use_input :
             text = self.readMessage ()
             if self.follow : self.follow.errorDirective (text)
             self.error (text)

       elif self.isIdentifier ():
          self.error ("Unknown directive " + self.tokenText)
       else :
          self.error ("Directive name expected")

       if not entry_use_input :
          self.readMessage ()
          # while self.token != self.eos and self.token != self.end_of_line :
           #   self.simpleToken ()

       if next_name == None : # !?
          if self.token != self.end_of_line :
             self.warning ("End of line (after directive) expected")
             self.readMessage()
             # while self.token != 0 and self.token != self.end_of_line :
             #    self.simpleToken ()

       if not entry_use_input and not cond_directive:
          if self.follow : self.follow.unusedDirective ()

       self.inside_directive = False

       if next_name != None :
          full_name = self.searchFile (next_name)
          if full_name == None :
             if self.tolerate_missing_include :
                self.warning ("Include file not found " + next_name)
             else :
                self.error ("Include file not found " + next_name)
          if full_name != None :
             next_input = LexerInput ()
             next_input.above = self.input
             next_input.isInclude = True
             # next_input.lastChar = self.ch
             self.input_stack.append (self.input)
             self.input = next_input
             self.input.openFile (full_name)
             if self.follow: self.follow.openInclude ()
             self.nextChar ()
             self.expandToken ()
       else :
          self.expandToken ()

   def nextToken (self) :
       self.expandToken ()
       again = True
       while again :
          again = True
          if self.isSeparator ('#') :
             self.directive ()
          elif self.isEndOfSource () and len (self.input_stack) != 0 :
             # lastChar = self.input.lastChar
             if self.input.isInclude:
                if self.follow: self.follow.closeInclude ()
             if self.input.isMacro:
                if self.follow: self.follow.closeExpand ()
             self.input = self.input_stack.pop ()
             # self.ch = lastChar
             # if self.follow: self.follow.oneCharacter ()
             self.repeatChar ()
             # print ("CLOSE")
             # NO self.nextChar ()
             self.expandToken ()
          else :
             if not self.use_input :
                if self.follow : self.follow.unusedToken ()
                self.simpleToken ()
             else :
                again = False
       if self.follow : self.follow.oneToken ()

   # -----------------------------------------------------------------------

   def isEndOfSource (self) :
       return self.token == self.eos

   def isIdentifier (self) :
       return self.token == self.identifier

   def isNumber (self) :
       return self.token == self.number

   def isString (self) :
       return self.token == self.string

   def isSeparator (self, value) :
       return self.token == self.separator and self.tokenText == value

   def isKeyword (self, value) :
       return self.token == self.identifier and self.tokenText == value

   def readIdentifier (self) :
       if not self.isIdentifier () :
          self.error ("Identifier expected")
       value = self.tokenText
       self.nextToken ()
       return value

   def readNumber (self) :
       if not self.isNumber () :
          self.error ("Number expected")
       value = self.tokenText
       self.nextToken ()
       return value

   def readString (self) :
       if not self.isString () :
          self.error ("String expected")
       value = self.tokenText
       self.nextToken ()
       return value

   def checkSeparator (self, value) :
       if not self.isSeparator (value) :
          self.error (value + " expected")
       self.nextToken ()

   def checkKeyword (self, value) :
       if not self.isKeyword (value) :
          self.error (value + " expected")
       self.nextToken ()

# --------------------------------------------------------------------------

# fileNames = ["file with index 1", "file with index 2" ]

# lexer_keywords = { "if" : toke_code_for_if , "while" : token_code_for_while }

# lexer_separators = { "<" : Separators (value = code_for_less,
#                            items = { "=" : ... , "<" : ... }),
#                      ";" : Separators (value = code_for_semicolons, items = { }) }

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    lexer = Lexer ()
    lexer.openFile ("pas.g")

    if 0 :
       for i in range (4) :
          print (lexer.tokenText)
          lexer.nextToken ()

       print ("---")

       point = lexer.mark ()
       for i in range (4) :
          print (lexer.tokenText)
          lexer.nextToken ()

       print ("---")
       lexer.rewind (point)
       for i in range (4) :
          print (lexer.tokenText)
          lexer.nextToken ()

    while not lexer.isEndOfSource () :
       print (lexer.tokenText, end=" ")
       lexer.nextToken ()

    point = lexer.mark ()
    lexer.rewind (point)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
