#!/usr/bin/env python

from __future__ import print_function

from grammar import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal, Directive
from info import incIndent, decIndent, put, putEol, putCondEol, putLn
from cgi import escape

# --------------------------------------------------------------------------

def putHRef (url, txt) :
    put ("<a href=" + '"' + url + '"' + ">" + txt + "</a>");

def putAName (name, txt) :
    put ("<a name=" + '"' + name + '"' + ">" + txt + "</a>");

def putHtmlBegin () :
    putLn ("<!DOCTYPE html>")
    putLn ("<html>")
    incIndent ()

    putLn ("<head>")
    incIndent ()
    putLn ("<title></title>")
    decIndent ()
    putLn ("</head>")
    putLn ("<body>")
    incIndent ()

    putLn ("<pre>")

def putHtmlEnd () :
    putLn ("</pre>")
    decIndent ()
    putLn ("</body>")
    decIndent ()
    putLn ("</html>")

# --------------------------------------------------------------------------

def htmlFromRules (grammar) :
    putHtmlBegin ()
    for rule in grammar.rules :
        htmlFromRule (grammar, rule)
    putHtmlEnd ()

def htmlFromRule (grammar, rule) :
    putAName (rule.name, rule.name)
    put (" ")
    put (":")
    putEol ()
    incIndent ()
    htmlFromExpression (grammar, rule.expr)
    putCondEol ()
    decIndent ()
    put (";")
    putEol ()
    putEol ()

def htmlFromExpression (grammar, expr) :
    n = 0;
    for alt in expr.alternatives :
       n = n + 1
       if n > 1 :
          put ("|")
          putEol ()
       htmlFromAlternative (grammar, alt)

def htmlFromAlternative (grammar, alt) :
    for item in alt.items :
        if isinstance (item, Terminal) :
           htmlFromTerminal (grammar, item)
        elif isinstance (item, Nonterminal) :
           htmlFromNonterminal (grammar, item)
        elif isinstance (item, Ebnf) :
           htmlFromEbnf (grammar, item)
        elif isinstance (item, Directive) :
           pass
        else :
           raise Exception ("Unknown alternative item")

def htmlFromEbnf (grammar, ebnf) :
    putCondEol ()
    put ("(")
    putEol ()
    incIndent ()
    htmlFromExpression (grammar, ebnf.expr)
    putEol ()
    decIndent ()
    put (")" + ebnf.mark)
    putEol ()
    if ebnf.impl != None :
       put ("=>")
       putEol ()
       htmlFromAlternative (grammar, ebnf.impl)

def htmlFromNonterminal (grammar, item) :
    name = item.rule
    putHRef ("#" + name, name)

def htmlFromTerminal (grammar, item) :
    name = item.text
    name = '"' + name + '"'
    put (escape (name))

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    grammar = Grammar ()
    grammar.openFile ("pas.g")
    grammar.parseRules ()
    htmlFromRules (grammar)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
