#!/usr/bin/env python

from __future__ import print_function

from grammar import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal, Directive
from info import incIndent, decIndent, put, putEol, putCondEol, putLn

# --------------------------------------------------------------------------

def loutFromRules (grammar) :
    putLn ("@SysInclude { tbl }")
    putLn ("@SysInclude { diag }")
    putLn ("@SysInclude { xrgb }") # colors
    putLn ("@SysInclude { doc }")
    putLn ("@Doc @Text @Begin")
    putLn ("")
    for rule in grammar.rules :
        loutFromRule (grammar, rule)
    putLn ("@End @Text")

def loutFromRule (grammar, rule) :
    putLn ("@SyntaxDiag")
    putLn ("title {" + rule.name + "}")
    putLn ("{")
    incIndent ()

    #count = 0
    #for alt in rule.expr.alternatives :
       #count = count + len (alt.items)
    #if count >= 8 :
       #putLn ("@StartDown")
    #else :
       #putLn ("@StartRight")
    putLn ("@StartRight")

    loutFromExpression (grammar, rule.expr)

    decIndent ()
    putLn ("}")
    putLn ("@PP")
    putLn ("@PP")
    putLn ("")

def loutFromExpression (grammar, expr) :
    count = len (expr.alternatives)
    if count == 0 :
       putLn ("@Skip")
    if count > 1 :
       putLn ("@Select")

    inx = 0
    groups = 0 # groups of 10 items
    for alt in expr.alternatives :
        inx = inx + 1
        if count > 1 :
           put (nodeName (inx) + " {")
           incIndent ()
           if inx == 10 and count > 10 : # maximum is 10 items
              putEol ()
              putLn ("@Select")
              inx = 1 # count again from 1
              groups = groups + 1
              put (nodeName (inx) + " {")
        loutFromAlternative (grammar, alt)
        if count > 1 :
           decIndent ()
           putLn ("}")

    while groups > 0 :
       decIndent ()
       putLn ("}")
       groups = groups - 1

def nodeName (inx) :
    if inx <= 26 :
        name = chr (ord ('A') + inx - 1)
    else :
        name = 'Z' + str (inx)
    return name

def loutFromAlternative (grammar, alt) :
    count = len (alt.items)
    if count == 0 :
       putLn ("@Skip")
    if count > 1 :
       putLn ("@Sequence")

    inx = 0
    groups = 0
    for item in alt.items :
        inx = inx + 1
        if count > 1 :
           put (nodeName (inx) + " {")
           incIndent ()
           if inx == 10 and count > 10 : # maximum is 10 items
              putEol ()
              putLn ("@Sequence")
              inx = 1 # count again from 1
              groups = groups + 1
              put (nodeName (inx) + " {")
        if isinstance (item, Terminal) :
           loutFromTerminal (grammar, item)
        elif isinstance (item, Nonterminal) :
           loutFromNonterminal (grammar, item)
        elif isinstance (item, Ebnf) :
           loutFromEbnf (grammar, item)
        elif isinstance (item, Directive) :
           pass
        else :
           raise Exception ("Unknown alternative item")
        if count > 1 :
           decIndent ()
           putLn ("}")

    while groups > 0 :
       decIndent ()
       putLn ("}")
       groups = groups - 1

def loutFromEbnf (grammar, ebnf) :
    if ebnf.mark == '?' :
       putLn ("@Optional")
    if ebnf.mark == '*' :
       putLn ("@Optional")
       putLn ("@RepeatOpposite")
    if ebnf.mark == '+' :
       putLn ("@Optional")
       putLn ("@RepeatOpposite")

    if ebnf.impl == None :
       incIndent ()
       loutFromExpression (grammar, ebnf.expr)
       decIndent ()
    else :
       # do not print syntactic predicate
       # putLn ("@Skip")
       loutFromAlternative (grammar, ebnf.impl)

def loutFromNonterminal (grammar, item) :
    name = item.rule
    put ("@ACell " + '"' + name + '"')

def loutFromTerminal (grammar, item) :
    value = item.text
    value = '"' + value + '"' # !?
    if len (value) == 1 :
       value = "@CCell " + value
    else :
       value = "@BCell " + value
    value = "blue @Color " + value
    put (value)

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    grammar = Grammar ()
    grammar.openFile ("pas.g")
    grammar.parseRules ()
    loutFromRules (grammar)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
