#!/usr/bin/env python

from __future__ import print_function

from grammar import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal, Assign, Style
from info import incIndent, decIndent, put, putEol, putCondEol, putLn
from symbols import Symbol, initSymbols

# --------------------------------------------------------------------------

def addToDictionary (tree, name) :
    ref = tree
    for c in name :
        if c not in ref :
            ref[c] = { }
        ref = ref[c]

def printPackedItem (dictionary, inx, symbol) :
    incIndent ()
    if len (dictionary) == 0 :
       putLn ("return " + '"' + symbol + '"' + ";")
    else :
       printPackedDictionary (dictionary, inx+1, symbol)
    decIndent ()

def printPackedDictionary (dictionary, inx, symbol) :
    if len (dictionary) == 1 :
       # put ("if ")
       # start = True
       start_inx = inx
       substr = ""
       while len (dictionary) == 1:
          # if not start :
          #    put (" and ")
          # start = False
          for c in sorted (dictionary.keys ()) : # only one
              # put ("s[" + str (inx) + "] == " + "'" + c + "'")
              substr = substr + c
              dictionary = dictionary [c]
              inx = inx + 1
              symbol = symbol + c
       # putLn (" :")
       inx = inx - 1
       if start_inx == inx :
          putLn ("if s[" + str (inx) + "] == " + "'" + substr + "'" + " :")
       else :
          putLn ("if s[" + str (start_inx) + ":" + str (inx) + "] == " + '"' + substr + '"' + " :")
       printPackedItem (dictionary, inx, symbol)
    else :
        for c in sorted (dictionary.keys ()) :
            putLn ("if s[" + str (inx) + "] == "  "'" + c + "'" + " : ")
            printPackedItem (dictionary[c], inx, symbol+c)

def printDictionary (dictionary) :
    for c in sorted (dictionary.keys ()) :
        putLn ("'" + c + "'" + " :")
        incIndent ()
        if len (dictionary[c]) == 0 :
           putLn ("#")
        else :
           printDictionary (dictionary[c])
        decIndent ()

def selectKeywords (grammar) :
    lim = 0
    for symbol in grammar.keywords :
        n = len (symbol)
        if n > lim :
           lim = n

    size = 1
    while size <= lim :
       tree = { }
       for symbol in grammar.keywords :
           if len (symbol) == size :
              addToDictionary (tree, symbol)
       if len (tree) != 0 :
          putLn ("if len (s) == " + str (size) + " :")
          incIndent ()
          printPackedDictionary (tree, 0, "")
          decIndent ()
       size = size + 1

    # tree = { }
    # for symbol in grammar.keywords :
    #     addToDictionary (tree, symbol)
    # printDictionary (tree)

    tree = { }
    for symbol in grammar.delimiters :
        addToDictionary (tree, symbol)
    printDictionary (tree)

# --------------------------------------------------------------------------

def initCollections (grammar) :
    grammar.collections = [ ]

def registerCollection (grammar, collection) :
    data = [ ]
    for inx in range (len (collection)) :
        if collection [inx] :
           data = data + [inx]
    if data not in grammar.collections :
       grammar.collections = grammar.collections + [data]
    return grammar.collections.index (data)

def printCollections (grammar) :
    num = 0
    for data in grammar.collections :
        txt = ""
        for inx in data :
            txt = txt + " " + grammar.symbols[inx].name

        putLn ("set" + str (num) + " = " + str (data) + " #" + txt)
        num = num + 1

def condition (grammar, collection) :
    cnt = 0
    for inx in range (len (collection)) :
        if collection [inx] :
           cnt = cnt + 1

    if cnt == 0 :
       # grammar.error ("Empty set")
       return "nothing"
    elif cnt <= 3 :
       code = ""
       start = True
       for inx in range (len (collection)) :
           if collection [inx] :
              if not start :
                 code = code + " or "
              start = False
              code = code + "sy == " + '"' + grammar.symbols[inx].name + '"'
       return code
    else :
       num = registerCollection (grammar, collection)
       return "set" + str (num) + " [sy]";

# --------------------------------------------------------------------------

def parserFromRules (grammar) :
    for rule in grammar.rules :
        parserFromRule (grammar, rule)

def parserFromRule (grammar, rule) :
    params = ""
    if rule.rule_mode == "modify" :
       params = "result"
    putLn ("def parse_" + rule.name + " (" + params + ") :")

    incIndent ()
    if rule.rule_mode == "new" or rule.rule_mode == "" :
       putLn ("result = " + rule.rule_type + " ()")
    parserFromExpression (grammar, rule.expr)
    putLn ("return result")
    decIndent ()
    putEol ()

def parserFromExpression (grammar, expr) :
    cnt = len (expr.alternatives)
    start = True
    for alt in expr.alternatives :
        if cnt > 1 :
           cond = condition (grammar, alt.first)
           if start :
              put ("if")
           else :
              put ("elif")
           start = False
           putLn (" " + cond + " :")
           incIndent ()
        parserFromAlternative (grammar, alt)
        if cnt > 1 :
           decIndent ()

def parserFromAlternative (grammar, alt) :
    for item in alt.items :
        if isinstance (item, Terminal) :
           parserFromTerminal (grammar, item)
        elif isinstance (item, Nonterminal) :
           parserFromNonterminal (grammar, item)
        elif isinstance (item, Ebnf) :
           parserFromEbnf (grammar, item)
        elif isinstance (item, Assign) :
           putLn ("result." + item.variable + " = " + item.value)
        elif isinstance (item, Style) :
           pass
        else :
           raise Exception ("Unknown alternative item: " + item.__class__.__name__)

def parserFromEbnf (grammar, ebnf) :
    if ebnf.mark == '?' :
       put ("if ")
    elif ebnf.mark == '*' :
       put ("while ")
    elif ebnf.mark == '+' :
       put ("while ")

    if ebnf.mark != "" :
       cond = condition (grammar, ebnf.expr.first)
       put (cond)
       putLn (" :")
       incIndent ()

    parserFromExpression (grammar, ebnf.expr)

    if ebnf.mark != "" :
       decIndent ()

def parserFromNonterminal (grammar, item) :
    if item.add :
       put ("result.items.append (")
    if item.variable != "" :
       put ("result." + item.variable + " = ")

    put ("parse_" + item.rule)

    put (" (")
    if item.call :
       put ("result")
    put (")")
    if item.add :
       put (")")
    putLn ("")

def parserFromTerminal (grammar, item) :
    name = item.text
    putLn ("check (" + '"' + name + '"' + ")")

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    grammar = Grammar ()
    grammar.openFile ("pas.g")
    grammar.parseRules ()
    initSymbols (grammar)

    initCollections (grammar)
    parserFromRules (grammar)
    printCollections (grammar)
    selectKeywords (grammar)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
