#!/usr/bin/env python

from __future__ import print_function

from grammar import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal, Directive, Style
from info import incIndent, decIndent, put, putEol, putCondEol, putLn

# --------------------------------------------------------------------------

def productFromRules (grammar) :
    for rule in grammar.rules :
        productFromRule (grammar, rule)

def productFromRule (grammar, rule) :
    putLn ("def write_" + rule.name + " () :")
    incIndent ()
    productFromExpression (grammar, rule.expr)
    decIndent ()
    putEol ()

def productFromExpression (grammar, expr) :
    n = len (expr.alternatives)
    for alt in expr.alternatives :
        if n > 1 :
           putLn ("if :")
           incIndent ()
        productFromAlternative (grammar, alt)
        if n > 1 :
           putLn ("if :")
           decIndent ()

def productFromAlternative (grammar, alt) :
    for item in alt.items :
        if isinstance (item, Terminal) :
           productFromTerminal (grammar, item)
        elif isinstance (item, Nonterminal) :
           productFromNonterminal (grammar, item)
        elif isinstance (item, Ebnf) :
           productFromEbnf (grammar, item)
        elif isinstance (item, Style) :
           productFromStyle (grammar, item)
        elif isinstance (item, Directive) :
           pass
        else :
           raise Exception ("Unknown alternative item")

def productFromEbnf (grammar, ebnf) :
    if ebnf.mark == '?' :
       putLn ("if :")
    elif ebnf.mark == '*' :
       putLn ("while :")
    elif ebnf.mark == '+' :
       putLn ("while :")

    if ebnf.mark != "" :
       incIndent ()

    productFromExpression (grammar, ebnf.expr)

    if ebnf.mark != "" :
       decIndent ()

def productFromNonterminal (grammar, item) :
    putLn ("write_" + item.rule + " ()")

def productFromTerminal (grammar, item) :
    putLn ("put (" + '"' + item.text + '"' + ")")

def productFromStyle (grammar, item) :
    putLn (item.name + " ()")

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    grammar = Grammar ()
    grammar.openFile ("pas.g")
    grammar.parseRules ()
    productFromRules (grammar)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
