#!/usr/bin/env python

from __future__ import print_function

import sys, os, re, traceback

from util import use_qt5
if use_qt5 :
   from PyQt5.QtCore import *
   from PyQt5.QtGui import *
   from PyQt5.QtWidgets import *
else :
   from PyQt4.QtCore import *
   from PyQt4.QtGui import *

# --------------------------------------------------------------------------

class Redirect (object) :
   def __init__ (self, target) :
       self.target = target
       self.buffer = ""

       self.save_out = sys.stdout
       self.save_err = sys.stderr

       sys.stdout = self
       sys.stderr = self

   def write (self, text) :
       if self.save_out != None :
          self.save_out.write (text)

       try:
          text = self.buffer + text
          inx = text.find ("\n")
          while inx > 0 :
             line = text [ : inx ]
             self.target.buildOneLine (line)
             text = text [ inx + 1 : ]
             inx = text.find ("\n")

          self.buffer = text
       except :
           sys.stdout = self.save_out
           sys.stderr = self.save_err

           sys.stderr.write ("\n")
           sys.stderr.write ("EXCEPTION during stderr redirect")
           sys.stderr.write ("\n")
           sys.stderr.write ("\n")
           traceback.print_exc ()
           sys.stderr.write ("\n")

           sys.stdout = self
           sys.stderr = self

   def close (self) :
       if self.buffer != "" :
          self.target.append (self.buffer)
       sys.stdout = self.save_out
       sys.stderr = self.save_err

# --------------------------------------------------------------------------

class Output (QTextEdit) :

   def __init__ (self, parent=None) :
       super (Output, self).__init__ (parent)
       self.setLineWrapMode (QTextEdit.NoWrap)

       self.win = None
       self.stopAction = None
       self.locationProperty = QTextFormat.UserProperty
       self.directory = ""
       self.redirect = None
       self.redirect_count = 0

       self.red = Qt.red
       self.green = Qt.green
       self.blue = Qt.blue
       self.gray = Qt.gray
       self.norm = QColor (0, 0, 0)
       self.brown = QColor ("brown")
       self.orange = QColor ("orange")
       self.cornflowerblue = QColor ("cornflowerblue")

   # -- redirect output --

   def redirectOutput (self) :
       self.redirect_count = self.redirect_count + 1
       if self.redirect_count == 1 :
          self.redirect = Redirect (self)

   def restoreOutput (self) :
       self.redirect_count = self.redirect_count - 1
       if self.redirect_count == 0 :
          self.redirect.close ()

   # -- color output --

   def colorText (self) :
       cursor = self.textCursor ()

       format = cursor.charFormat ()
       format.setForeground (self.blue)
       cursor.setCharFormat (format)

       cursor.insertText ("blue line\n")

       format = cursor.charFormat ()
       format.setForeground (self.red)
       cursor.setCharFormat (format)

       cursor.insertText ("red line")

   def setColor (self, cursor, color) :
       format = cursor.charFormat ()
       format.setForeground (color)
       cursor.setCharFormat (format)

   def setNote (self, cursor, note) :
       format = cursor.charFormat ()
       format.setProperty (self.locationProperty, note)
       format.setToolTip (note)
       cursor.setCharFormat (format)

   def clearNote (self, cursor) :
       format = cursor.charFormat ()
       format.clearProperty (self.locationProperty)
       format.clearProperty (QTextFormat.TextToolTip)
       cursor.setCharFormat (format)

   # -- grep with ANSI colors --

   def colorDataReady (self) :
       cursor = self.textCursor ()
       cursor.movePosition (cursor.End)

       txt = ""
       esc = False
       code = ""
       normal = True
       for c in self.process.readAll () :
          if c == chr (27) :
             esc = True
             code = ""
          elif esc :
             code = code + c
             if not (c == '[' or c >= '0' and c <= '9' or c == ';') :
                if txt != "" :
                   cursor.insertText (txt)
                   txt = ""
                esc = False
                if code == "[35m" :
                   self.setColor (cursor, self.blue) # magenta
                   normal = False
                elif code == "[36m" :
                   self.setColor (cursor, self.gray)
                   normal = False
                elif code == "[32m" :
                   self.setColor (cursor, self.green)
                   normal = False
                elif code == "[01;31m" :
                   self.setColor (cursor, self.red)
                   normal = False
                elif code == "[m" :
                   if not normal :
                      self.setColor (cursor, self.norm)
                      normal = True
          else :
              txt = txt + c

       if txt != "" :
          cursor.insertText (txt)

       self.ensureCursorVisible ()

   # -- grep in output window --

   def grepPutLine (self, cursor, text) :
       self.setNote (cursor, text[1] + ":" + text[3])
       cursor.insertText (text [0])
       self.setColor (cursor, self.blue)
       cursor.insertText (text [1])
       self.setColor (cursor, self.gray)
       cursor.insertText (text [2])
       self.setColor (cursor, self.green)
       cursor.insertText (text [3])
       self.setColor (cursor, self.gray)
       cursor.insertText (text [4])
       self.setColor (cursor, self.norm)
       cursor.insertText (text [5])
       self.setColor (cursor, self.red)
       cursor.insertText (text [6])
       self.setColor (cursor, self.norm)
       cursor.insertText (text [7])

   def grepDataReady (self) :
       cursor = self.textCursor ()
       cursor.movePosition (QTextCursor.End)

       esc = False
       code = ""
       txt = ""
       section = 0
       text = ["", "", "", "", "", "", "", ""]
       for c in self.process.readAll () :
          if c == chr (10) :
             self.grepPutLine (cursor, text)
             cursor.insertText (c)
             text = ["", "", "", "", "", "", "", ""]
          elif c == chr (27) :
             esc = True
             code = ""
          elif esc :
             code = code + c
             if not (c == '[' or c >= '0' and c <= '9' or c == ';') :
                if txt != "" :
                   text [section] = text [section] + txt
                   txt = ""
                esc = False
                if code == "[35m" :
                   # magenta (blue)
                   section = 1
                elif code == "[36m" :
                   # gray
                   if section == 1 or section == 3 :
                      section = section + 1
                elif code == "[32m" :
                   # green
                   section = 3
                elif code == "[01;31m" :
                   # red
                   section = 6
                elif code == "[m" :
                   # black
                   if section == 4 or section == 6 :
                      section = section + 1
          else :
              txt = txt + c

       self.grepPutLine (cursor, text)
       self.ensureCursorVisible ()

   def grep (self, params) :
       self.process = QProcess (self)
       self.process.setProcessChannelMode (QProcess.MergedChannels)
       self.process.readyRead.connect (self.grepDataReady)
       if self.stopAction != None :
          self.process.started.connect (lambda: self.stopAction.setEnabled (True))
          self.process.finished.connect (lambda: self.stopAction.setEnabled (False))
       self.process.start ("/bin/sh", [ "-c", "grep " + params + " -n -r . --color=always" ] )
       # self.process.waitForFinished ()

   # -- gcc options --

   def gccLine (self, cursor, line) :
       for word in line.split () :
           style = False
           if word.startswith ("-I") :
              self.setColor (cursor, self.green)
              style = True
           elif word.startswith ("-D") or word.startswith ("-U") :
              self.setColor (cursor, self.orange)
              style = True
           elif word.startswith ("-L") :
              self.setColor (cursor, self.cornflowerblue)
              style = True
           elif word.startswith ("-l") :
              self.setColor (cursor, self.blue)
              style = True
           cursor.insertText (word)
           if style :
              self.setColor (cursor, self.norm)
           cursor.insertText (" ")

   # -- make output --

   def modifyLine (self, cursor, line, keyword) :
       pattern = "([^:]*):([^:]*):(.*)(" + keyword + ":)(.*)"
       m = re.match (pattern, line)
       if m :
          mark = os.path.join (self.directory, m.group (1)) + ":" + m.group (2) + ":" + m.group (3)

          self.setNote (cursor, mark)

          self.setColor (cursor, self.blue)
          cursor.insertText (m.group (1))
          self.setColor (cursor, self.norm)
          cursor.insertText (":")

          self.setColor (cursor, self.green)
          cursor.insertText (m.group (2))
          self.setColor (cursor, self.norm)
          cursor.insertText (":")

          self.setColor (cursor, self.gray)
          cursor.insertText (m.group (3))

          self.setColor (cursor, self.red)
          cursor.insertText (m.group (4))

          self.setColor (cursor, self.orange)
          cursor.insertText (m.group (5))
          self.setColor (cursor, self.norm)

          self.clearNote (cursor)  # before end of line, otherwise format is not changed
       else :
          self.setColor (cursor, self.brown)
          cursor.insertText (line[:-1])
          self.setColor (cursor, self.norm)

   def directoryLine (self, cursor, line) :
       pattern = "Entering directory '([^']*)'"
       m = re.search (pattern, line)
       if m :
          self.directory = m.group (1)
          self.setColor (cursor, self.orange)
       else :
          self.setColor (cursor, self.brown)
       cursor.insertText (line[:-1])
       self.setColor (cursor, self.norm)

   # -- Python trace --

   def pythonLine (self, cursor, line) :
       pattern = "  File (.*), line (.*), in (.*)"
       m = re.match (pattern, line)
       if m :
          fileName = m.group (1)
          line =  m.group (2)
          if fileName.startswith ('"') and fileName.endswith ('"') :
             fileName = fileName [ 1 : -1 ]
          mark = fileName + ":" + line
          self.setNote (cursor, mark)

          cursor.insertText ("File ")
          self.setColor (cursor, self.blue)
          cursor.insertText (m.group (1))
          self.setColor (cursor, self.norm)

          cursor.insertText (", line ")
          self.setColor (cursor, self.green)
          cursor.insertText (m.group (2))
          self.setColor (cursor, self.norm)

          cursor.insertText (", in ")
          self.setColor (cursor, self.orange)
          cursor.insertText (m.group (3))
          self.setColor (cursor, self.norm)

          self.clearNote (cursor) # before end of line, otherwise format is not changed
       else :
          self.setColor (cursor, self.brown)
          cursor.insertText (line)
          self.setColor (cursor, self.norm)

   def buildOneLine (self, txt) :
       cursor = self.textCursor ()
       cursor.movePosition (QTextCursor.End)
       if txt.find ("error:") >= 0 :
          self.modifyLine (cursor, txt, "error")
       elif txt.find ("warning:") >= 0 :
          self.modifyLine (cursor, txt, "warning")
       elif txt.find ("Entering directory ") >= 0 :
          self.directoryLine (cursor, txt)
       elif txt.find ("Leaving directory ") >= 0 :
          self.directory = ""
       elif txt.startswith ("  File ") :
          self.pythonLine (cursor, txt)
       elif txt.find ("gcc") >= 0 or txt.find ("g++") >= 0 :
          self.gccLine (cursor, txt)
       else :
          cursor.insertText (txt)
       cursor.insertText ("\n")

   def buildDataReady (self) :
       cursor = self.textCursor ()
       cursor.movePosition (cursor.End)

       txt = ""
       for c in self.process.readAll () :
           txt = txt + c
           if c == '\n' :
              self.buildOneLine (txt)
              txt = ""
       if txt != "" :
          self.buildOneLine (txt)
          self.ensureCursorVisible ()

   # -- run command --

   def readEnvironment (self, group) :
       self.win.settings.beginGroup (group)
       env_name = self.win.stringValue ("env");
       self.win.settings.endGroup ()

       if env_name != "" :
          self.readEnvironment (env_name); # read other environment

       self.win.settings.beginGroup (group)

       work_dir = self.win.stringValue ("cd")
       if work_dir != "" :
          self.process.setWorkingDirectory (work_dir)
          # print ("DIRECTORY", work_dir)

       env = QProcessEnvironment.systemEnvironment () # important
       keys = self.win.settings.allKeys ()
       for key in keys :
          if key.startsWith ("set") or key.startsWith ("var") :
             text = str (self.win.stringValue (key))
             answer = text.split ("=", 1)
             name = answer [0]
             value = answer [1]
             name = name.strip ()
             value = value.strip ()
             env.insert (name, value)
             # print ("SET", name + "=" + value)
       self.process.setProcessEnvironment (env)

       self.win.settings.endGroup ()

   def run (self, group) :

       self.directory = ""
       self.process = QProcess (self)
       self.process.setProcessChannelMode (QProcess.MergedChannels)
       self.process.readyRead.connect (self.buildDataReady)
       if self.stopAction != None :
          self.process.started.connect (lambda: self.stopAction.setEnabled (True))
          self.process.finished.connect (lambda: self.stopAction.setEnabled (False))

       self.win.settings.beginGroup (group)
       cmd = self.win.stringValue ("cmd");
       self.win.settings.endGroup ()

       self.readEnvironment (group)

       self.process.start ("/bin/sh", [ "-c", cmd ] )

   def stop (self) :
       self.process.terminate ()

   def clearOutput (self) :
       self.clear ()
       # cursor = self.textCursor ()
       # cursor.movePosition (QTextCursor.Start)

   # -- mouse click --

   def mousePressEvent (self, e) :
       cursor = self.cursorForPosition (e.pos ())
       cursor.select (QTextCursor.WordUnderCursor)
       format = cursor.charFormat ()
       mark = format.stringProperty (self.locationProperty)
       if mark != "" :
          # QToolTip.showText (e.globalPos (), mark)
          self.goToLocation (mark)
       # else :
       #    QToolTip.hideText ()
       super (Output, self).mouseMoveEvent (e)

   def goToLocation (self, mark) :
       pattern = "([^:]*)(:([0-9]*)(:([0-9]*))?)?"
       m = re.match (pattern, mark)
       if m :
          fileName = m.group (1)
          line = m.group (3)
          column = m.group (5)
          if line == None or line == "" :
             line = 0
          else :
             line = int (line)
          if column == None or column == "" :
             column = 0
          else :
             column = int (column)
          # print ("file:", fileName)
          # print ("line:",  line)
          # print ("column:", column)
          if self.win != "" :
             self.win.loadFile (fileName, line, column)

   # -- jump to next mark --

   def jumpToPrevMark (self) :
       cursor = self.textCursor ()
       cursor.movePosition (QTextCursor.StartOfLine)
       stop = False
       found = False
       while not stop and not found:
          if not cursor.movePosition (QTextCursor.Up) :
             stop = True
          if cursor.atStart () :
             stop = True
          if cursor.charFormat ().hasProperty (self.locationProperty) :
             found = True

       if found :
          cursor.movePosition (QTextCursor.EndOfLine, QTextCursor.KeepAnchor)

       self.setTextCursor (cursor)
       self.ensureCursorVisible ()

       if found :
          mark = cursor.charFormat ().stringProperty (self.locationProperty)
          self.goToLocation (mark)
          self.win.status.showMessage ("")
       else :
          self.win.status.showMessage ("Begin of file")

   def jumpToNextMark (self) :
       cursor = self.textCursor ()
       cursor.movePosition (QTextCursor.StartOfLine)
       stop = False
       found = False
       while not stop and not found:
          if not cursor.movePosition (QTextCursor.Down) :
             stop = True
          if cursor.atEnd () :
             stop = True
          if cursor.charFormat ().hasProperty (self.locationProperty) :
             found = True

       if found :
          cursor.movePosition (QTextCursor.EndOfLine, QTextCursor.KeepAnchor)

       self.setTextCursor (cursor)
       self.ensureCursorVisible ()

       if found :
          mark = cursor.charFormat ().stringProperty (self.locationProperty)
          self.goToLocation (mark)
          self.win.status.showMessage ("")
       else :
          self.win.status.showMessage ("End of file")

# http://stackoverflow.com/questions/26500429/qtextedit-and-colored-bash-like-output-emulation
# http://stackoverflow.com/questions/22069321/realtime-output-from-a-subprogram-to-stdout-of-a-pyqt-widget

# --------------------------------------------------------------------------

if __name__ == '__main__' :
   app = QApplication (sys.argv)
   win = Output ()
   win.show ()
   win.grep ("class")
   app.exec_ ()

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
