
# simple.py

from __future__ import print_function

import sys, os

from util import use_qt5

if use_qt5 :
   from PyQt5.QtCore import *
   from PyQt5.QtGui import *
   from PyQt5.QtWidgets import *
else :
   from PyQt4.QtCore import *
   from PyQt4.QtGui import *

sys.path.insert (1, os.path.join (os.getcwd (), "../grm")) # directory with parser

from lexer import Lexer

from area import Area
from edit import Editor

# --------------------------------------------------------------------------

class Simple (object) :
   def __init__ (self) :
       self.ink = None
       self.paper = None
       self.tooltip = None

class Module (Simple) :
   def __init__ (self) :
       super (Module, self).__init__ ()
       self.name = ""
       self.items = [ ]

class Class (Simple) :
   def __init__ (self) :
       super (Class, self).__init__ ()
       self.name = ""
       self.items = [ ]

class Function (Simple) :
   def __init__ (self) :
       super (Function, self).__init__ ()
       self.name = ""
       self.type = ""
       self.parameters = [ ]

class Variable (Simple) :
   def __init__ (self) :
       super (Variable, self).__init__ ()
       self.name = ""
       self.type = ""
       self.value = ""
       self.items = [ ]

# --------------------------------------------------------------------------

class SimpleLexer (Lexer) :

   infoProperty = QTextFormat.UserProperty

   def __init__ (self) :
       super (SimpleLexer, self).__init__ ()
       self.cursor = None
       self.blockMode = False
       self.path = [ ]

   def setEditor (self, editor) :
       self.cursor = editor.edit.textCursor ()

   def startPos (self) :
       return self.tokenByteOfs

   def stopPos (self) :
       return self.byteOfs

   def setBlock (self, start, stop) :
       self.cursor.setPosition (start)
       self.cursor.setPosition (stop, QTextCursor.KeepAnchor)
       self.blockMode = True

   def releaseBlock (self) :
       self.blockMode = False

   def moveCursor (self) :
       if not self.blockMode :
          self.cursor.setPosition (self.tokenByteOfs)
          self.cursor.setPosition (self.tokenByteOfs + len (self.tokenText), QTextCursor.KeepAnchor)

       # self.cursor.movePosition (QTextCursor.Start)
       # if self.tokenLineNum > 1 :
       #    self.cursor.movePosition (QTextCursor.Down, QTextCursor.MoveAnchor, self.tokenLineNum-1)
       # self.cursor.movePosition (QTextCursor.StartOfLine)
       # if self.tokenColNum > 1 :
       #    self.cursor.movePosition (QTextCursor.Right, QTextCursor.MoveAnchor, self.tokenColNum-1)
       # self.cursor.movePosition (QTextCursor.Right, QTextCursor.KeepAnchor, len (self.tokenText))

   # set character format

   def setForeground (self, brush) :
       self.moveCursor ()
       fmt = self.cursor.charFormat ()
       fmt.setForeground (brush)
       self.cursor.setCharFormat (fmt)

   def setBackground (self, brush) :
       self.moveCursor ()
       fmt = self.cursor.charFormat ()
       fmt.setBackground (brush)
       self.cursor.setCharFormat (fmt)

   def setFontItalic (self, italic = True) :
       self.moveCursor ()
       fmt = self.cursor.charFormat ()
       fmt.setFontItalic (italic)
       self.cursor.setCharFormat (fmt)

   def setUnderlineColor (self, color) :
       self.moveCursor ()
       fmt = self.cursor.charFormat ()
       fmt.setFontUnderline (True) # important
       fmt.setUnderlineColor (color)
       self.cursor.setCharFormat (fmt)

   def setToolTip (self, tooltip) :
       self.moveCursor ()
       fmt = self.cursor.charFormat ()
       fmt.setToolTip (tooltip)
       self.cursor.setCharFormat (fmt)

   def setInfo (self, info) :
       self.moveCursor ()
       fmt = self.cursor.charFormat ()
       fmt.setProperty (self.infoProperty, info)
       self.cursor.setCharFormat (fmt)

   # lexer

   def getIdentifier (self, msg = "") :
       name = ""
       if not self.isIdentifier () :
          if msg == "" :
             msg = "Identifier"
          self.error (msg + " expected")
       else :
          name = self.tokenText
       return name

   def readIdentifier (self, msg = "") :
       name = self.getIdentifier (msg)
       self.nextToken ()
       return name

   def readString (self) :
       if self.token != self.string  :
          self.error ("String expected")
       value = self.tokenText
       self.nextToken ()
       return value

   def optionalSemicolon (self) :
       if self.isSeparator (";") :
          self.nextToken ()

   # scope

   def openScope (self, name) :
       self.path.append (name)

   def closeScope (self) :
       self.path.pop ()

   def scopeName (self) :
       return ".".join (self.path)

# --------------------------------------------------------------------------

class SimpleParser (SimpleLexer) :

   identProperty = QTextFormat.UserProperty

   def __init__ (self) :
       super (SimpleParser, self).__init__ ()
       self.fileName = ""

   def parseNamespace (self) :
       result = Module ()
       result.line = self.tokenLineNum
       color = self.findColor ("namespace")
       self.setForeground (color)
       self.nextToken ()

       result.name = self.getIdentifier ("Module name")
       self.openScope (result.name)
       self.setInfo (self.scopeName ())
       self.setToolTip ("namespace identifier")
       self.setForeground (color)
       self.nextToken ()

       self.setToolTip ("opening")
       self.setForeground (color)
       self.checkSeparator ("{")

       while not self.isSeparator ("}") :
          if not self.parseAttributes (result) :
             item = self.parseCode ()
             result.items.append (item)

       self.setToolTip ("closing")
       self.setForeground (color)
       self.checkSeparator ("}")
       self.optionalSemicolon ()

       self.closeScope ()
       return result

   def parseClass (self) :
       result = Class ()
       result.line = self.tokenLineNum
       color = self.findColor ("class")
       self.setForeground (color)
       # start = self.startPos ()
       self.nextToken ()

       result.name = self.getIdentifier ("Class name")
       self.openScope (result.name)
       self.setInfo (self.scopeName ())
       self.setForeground (color)
       self.nextToken ()

       self.setForeground (color)
       self.checkSeparator ("{")
       while not self.isSeparator ("}") :
          if not self.parseAttributes (result) :
             item = self.parseCode ()
             result.items.append (item)
       self.setForeground (color)
       # stop = self.stopPos ()
       self.checkSeparator ("}")
       self.optionalSemicolon ()

       self.closeScope ()
       # self.setBlock (start, stop)
       # if result.paper != None :
       #    self.setBackground (QColor (result.paper))
       # self.releaseBlock ()
       return result

   def parseParameter (self, func) :
       item = Variable ()
       # self.setUnderlineColor (Qt.blue)
       item.type = self.readIdentifier ("Parameter type")
       item.name = self.readIdentifier ("Parameter name")
       func.parameters.append (item)

   def parseFunction (self) :
       result = Function ()
       result.line = self.tokenLineNum
       color = self.findColor ("function")
       # start = self.startPos ()

       result.type = self.readIdentifier ("Function type")
       if self.isIdentifier () :
          result.name = self.readIdentifier ("Function name")
       else :
          result.name = result.type
          result.type = ""

       self.openScope (result.name)
       self.setInfo (self.scopeName ())

       self.setForeground (color)
       self.checkSeparator ("(")

       if not self.isSeparator (")") :
          self.parseParameter (result)
          while self.isSeparator (",") :
             self.nextToken ()
             self.parseParameter (result)

       self.setForeground (color)
       self.checkSeparator (")")

       # stop = self.stopPos ()
       if self.isSeparator ("{") :
          self.setForeground (color)
          self.checkSeparator ("{")

          level = 1
          while level > 0 :
             if not self.parseAttributes (result) :
                if self.isSeparator ("{") :
                   level = level + 1
                if self.isSeparator ("}") :
                   level = level - 1
                if level > 0 :
                   self.nextToken ()

          self.setForeground (color)
          # stop = self.stopPos ()
          self.checkSeparator ("}")

       self.optionalSemicolon ()
       self.closeScope ()

       # self.setBlock (start, stop)
       # if result.paper != None :
       #    self.setBackground (QColor (result.paper))
       # self.releaseBlock ()
       return result

   def parseVariable (self) :
       result = Variable ()
       result.line = self.tokenLineNum
       color = self.findColor ("variable")

       result.type = self.readIdentifier ("Variable type")
       if self.isIdentifier () :
          result.name = self.readIdentifier ("variable name")
       else :
          result.name = result.type
          result.type = ""

       if result.name != "" :
          self.openScope (result.name)
          self.setInfo (self.scopeName ())
       self.setForeground (color)

       if self.isSeparator ("{") :
          self.setForeground (color)
          self.checkSeparator ("{")

          while not self.isSeparator ("}") :
             if not self.parseAttributes (result) :
                item = self.parseCode ()
                result.items.append (item)

          self.setForeground (color)
          self.checkSeparator ("}")
          self.optionalSemicolon ()

       else :
          if self.isSeparator ("=") :
             self.nextToken ()
             if self.token == self.identifier or self.token == self.string :
                result.value = self.tokenText
                self.nextToken ()
             else :
                self.error ("Variable value expected")

          while not self.isSeparator (";") and not self.isSeparator ("}") and self.token != self.eos :
             self.nextToken ();

          self.checkSeparator (";")

       if result.name != "" :
          self.closeScope ()
       return result

   def parseAttributeWithParameter (self) :
       self.nextToken () # attribute name
       self.checkSeparator ("(")
       text = self.readString ()
       self.checkSeparator (")")
       self.optionalSemicolon ()
       return text

   def parseAttributes (self, target) :
       cont = True
       count = 0
       while cont :
          count = count + 1
          if self.isKeyword ("warning") :
             line = self.tokenLineNum
             col = self.tokenColNum
             text = self.parseAttributeWithParameter ()
             print (self.fileName + ":" + str (line) + ":" + str (col) + ": warning: " + text)
          elif self.isKeyword ("error") :
             line = self.tokenLineNum
             col = self.tokenColNum
             text = self.parseAttributeWithParameter ()
             print (self.fileName + ":" + str (line) + ":" + str (col) + ": error: " + text)
          elif self.isKeyword ("ink") :
             target.ink = self.parseAttributeWithParameter ()
          elif self.isKeyword ("paper") :
             target.paper = self.parseAttributeWithParameter ()
          elif self.isKeyword ("tooltip") :
             target.tooltip = self.parseAttributeWithParameter ()
          else :
             count = count - 1
             cont = False
       return count > 0

   def parseCode (self) :
       result = None

       if self.isKeyword ("namespace") :
          result = self.parseNamespace ()

       elif self.isKeyword ("class") :
          result = self.parseClass ()

       elif self.isIdentifier () :
          func = False
          pos = self.mark ()
          self.nextToken () # first identifier
          if self.isIdentifier () :
             self.nextToken () # next (optional) identifier
          if self.isSeparator ("(") :
             func = True
          self.rewind (pos)

          if func :
             result = self.parseFunction ()
          else:
             result = self.parseVariable ()

       else :
          self.error ("Declaration expected")

       return result

# --------------------------------------------------------------------------

class SimpleBrowser (object) :

    def __init__ (self, main_window, fileName, code) :
       super (SimpleBrowser, self).__init__ ()
       self.win = main_window
       self.fileName = fileName
       self.displayBranch (None, code)

    def displayBranch (self, branch, code) :

        color = None
        icon = None

        if isinstance (code, Module) :
           text = "namespace " + code.name
           color = "namespace"
           icon = "namespace"
        elif isinstance (code, Class) :
           text = "class " + code.name
           color = "class"
           icon = "class"
        elif isinstance (code, Function) :
           text = "function " + code.name
           color = "function"
           icon = "function"
        elif isinstance (code, Variable) :
           text = "variable " + str (code.name)
           if code.type != None :
              text = text + " : " + code.type
           color = "variable"
           icon = "variable"
        else :
           text = str (code)

        node = QTreeWidgetItem ()
        node.setText (0, text)

        if color != None :
           node.setForeground (0, self.win.findColor (color))

        if icon != None :
           icon = self.win.findIcon (icon)
           if icon != None :
              node.setIcon (0, icon)

        if code.ink != None :
            node.setForeground (0, QColor (code.ink))
        if code.paper != None :
            node.setBackground (0, QColor (code.paper))
        if code.tooltip != None :
            node.setToolTip (0, code.tooltip)

        self.win.tree.add (branch, node)

        if branch == None :
           node.fileName = self.fileName

        if hasattr (code, "line") :
           node.line = code.line

        node.object = code

        if isinstance (code, Module) or isinstance (code, Class)  or isinstance (code, Variable) :
           for item in code.items :
               self.displayBranch (node, item)
        elif isinstance (code, Function) :
           for item in code.parameters :
               self.displayBranch (node, item)

# --------------------------------------------------------------------------

class SimpleDesigner (object) :

    def __init__ (self, main_window, code) :
       super (SimpleDesigner, self).__init__ ()
       self.win = main_window
       self.designBranch (code, 100, 100)
       self.win.design.addArea (code.area)

    def designBranch (self, code, x0, y0) :
        if isinstance (code, Variable) and (code.type.startswith ("Q") or code.type == "" and code.name.startswith ("Q")):
           self.designWidget (code, x0, y0)
        else :
           self.designArea (code, x0, y0)

    def designArea (self, code, x0, y0) :
        area = Area ()
        area.setToolTip (code.name)

        if isinstance (code, Module) :
           color = "namespace"
        elif isinstance (code, Class) :
           color = "class"
        elif isinstance (code, Function) :
           color = "function"
        elif isinstance (code, Variable) :
           color = "variable"
        else :
           color = ""

        code.area = area

        color = self.win.findColor (color)
        area.setBrush (QBrush (color))
        area.setPen (QPen (Qt.red))

        if code.ink != None:
            area.setPen (QColor (code.ink))
        if code.paper != None :
            area.setBrush (QColor (code.paper))
        if code.tooltip != None :
            area.setToolTip (code.tooltip)

        xspace = 20
        yspace = 20
        x = x0 + xspace
        y = y0 + yspace
        w = 60 + 2 * xspace
        h = 2 * yspace
        if hasattr (code, "items") :
           for item in code.items :
               self.designBranch (item, x, y)

               rect = item.area.boundingRect ()
               width = rect.width ()
               height = rect.height ()

               area.addArea (item.area)
               if isinstance (item.area, QGraphicsProxyWidget) :
                  item.area.setPos (x, y + yspace)
               else :
                  item.area.setRect (x, y, width, height)

               if w < width + 2*xspace :
                  w = width + 2*xspace
               y = y + height + yspace
               h = h + height + yspace

        # rect = area.boundingRect ()
        area.setRect (x0, y0, w, h)

    def designWidget (self, code, x0, y0) :
        if code.type == "" :
           code.type = code.name
           code.name = ""

        widget = eval (code.type + " ()")

        proxy = QGraphicsProxyWidget ()
        proxy.setWidget (widget)
        # proxy.setWindowFlags (Qt.Dialog)
        proxy.setFlag (QGraphicsItem.ItemIsMovable)
        proxy.setFlag (QGraphicsItem.ItemIsSelectable)

        code.area = proxy
        self.win.design.addArea (proxy)

        for item in code.items :
           if isinstance (item, Variable) :
              if item.value != None :
                 # setattr (widget, item.name, item.value)
                 quote = '\"'
                 if item.type == "QColor" :
                     param = "QColor (" + quote + item.value + quote + ")"
                 else :
                     param = quote + item.value + quote
                 eval ("widget.set" + item.name.capitalize() + " (" + param + ")")

# --------------------------------------------------------------------------

class SimpleHighlighter (QSyntaxHighlighter) :
   def __init__ (self, parent = None) :
       super (SimpleHighlighter, self).__init__ (parent)
       self.lexer = Lexer ()

       self.moduleFormat = self.format (Qt.green)
       self.classFormat = self.format (Qt.blue)
       self.functionFormat = self.format (Qt.darkRed)
       # self.variableFormat = self.format (QColor ("orange"))

   def format (self, color) :
       result = QTextCharFormat ()
       result.setForeground (color)
       return result

   def highlightBlock (self, text) :
       # print "highlightBlock", text
       lexer = self.lexer
       lexer.openString (text)
       while not lexer.isEndOfSource () :
          index = lexer.tokenCharPos - 1
          length = lexer.charPos - lexer.tokenCharPos
          if lexer.token == lexer.identifier :
             if lexer.tokenText == "namespace" :
                self.setFormat (index, length, self.moduleFormat)
             elif lexer.tokenText == "class" :
                self.setFormat (index, length, self.classFormat)
             elif lexer.tokenText == "function" :
                self.setFormat (index, length, self.functionFormat)
          lexer.nextToken ()

# --------------------------------------------------------------------------

class SimpleCompleter (QCompleter) :
   def __init__ (self, parent=None) :
       words = [ "namespace", "class", "function" ]
       super (SimpleCompleter, self).__init__ (words, parent)

# --------------------------------------------------------------------------

# class ViewWindow
#    variables
#       tree : Tree
#    methods
#       getEditor () : Editor
#       hasExtension (editor: Editor, extension: string) : boolean

# class Tree : QTreeWidget
#    methods
#       add (branch, node)

# --------------------------------------------------------------------------

class SimplePlugin (QObject) :

   def __init__ (self, main_window) :
       super (SimplePlugin, self).__init__ (main_window)
       self.win = main_window

       self.simpleMenu = self.win.menuBar().addMenu ("&Simple")
       self.simpleMenu.aboutToShow.connect (self.onShowSimpleMenu)

       act = QAction ("&Compile", self.win)
       act.setShortcut ("F10")
       act.triggered.connect (self.simpleCompile)
       self.simpleMenu.addAction (act)

   def onShowSimpleMenu (self) :
       is_simple = self.win.hasExtension (self.win.getEditor(), ".sim")
       for act in self.simpleMenu.actions () :
          act.setEnabled (is_simple)

   def simpleCompile (self) :
       e = self.win.getEditor ()
       if self.win.hasExtension (e, ".sim") :
          source = str (e.edit.toPlainText ())
          e.edit.setPlainText (source)

          self.win.info.redirectOutput ()

          parser = SimpleParser ()
          parser.openString (source)
          parser.setEditor (e)
          parser.fileName = e.origFileName
          parser.findColor = self.win.findColor

          code = parser.parseCode ()
          print ("parsed")

          self.win.info.restoreOutput ()

          # e.edit.setHighlighter (SimpleHighlighter) # class name, not instance

          e.edit.setCompleter (SimpleCompleter ())

          SimpleBrowser (self.win, e.origFileName, code)

          SimpleDesigner (self.win, code)

          # editor = Editor ()
          # SimpleHighlighter (editor.edit.document ())
          # self.win.tabWidget.addTab (editor, "")
          # self.win.tabWidget.setCurrentWidget (editor)

# --------------------------------------------------------------------------

if __name__ == "__main__" :
    parser = SimpleParser ()
    parser.openFile ("example.sim")
    code = parser.parseCode ()

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
